# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op


"""add index to agents host

Revision ID: cd9ef14ccf87
Revises: 8160f7a9cebb
Create Date: 2022-01-07 15:45:33.319170

"""

# revision identifiers, used by Alembic.
revision = 'cd9ef14ccf87'
down_revision = '8160f7a9cebb'


TABLE = 'agents'
COLUMN = 'host'


def upgrade():
    op.create_index(op.f('ix_' + TABLE + '_' + COLUMN), TABLE, [COLUMN])
