# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pyroute2 import devlink

from neutron import privileged
from neutron.privileged.agent import linux as priv_linux


@privileged.default.entrypoint
def get_port_list():
    dl = None
    try:
        dl = devlink.DL()
        return priv_linux.make_serializable(dl.port_list())
    finally:
        if dl:
            dl.close()
